from enum import Enum


class BOIMMGProperties(Enum):
     FORMULA = "formula"
     NAME = "name"
     MASS = "mass"
     CHARGE = "charge"
     GENERIC = "generic"
     SMILES = "smiles"
     INCHI = "inchi"
     INCHIKEY = "inchikey"
